@@ -16,8 +16,14 @@ from awkward._backends.dispatch import (
 )
 from awkward._behavior import get_array_class, get_record_class
 from awkward._kernels import KernelError
-from awkward._layout import wrap_layout
+from awkward._layout import maybe_posaxis, wrap_layout
 from awkward._meta.meta import Meta
+from awkward._namedaxis import (
+    NamedAxis,
+    _add_named_axis,
+    _keep_named_axis,
+    _remove_named_axis,
+)
 from awkward._nplikes import to_nplike
 from awkward._nplikes.dispatch import nplike_of_obj
 from awkward._nplikes.numpy import Numpy
@@ -27,7 +33,12 @@ from awkward._parameters import (
     parameters_are_equal,
     type_parameters_equal,
 )
-from awkward._regularize import is_integer_like, is_sized_iterable
+from awkward._regularize import (
+    is_array_like,
+    is_integer,
+    is_integer_like,
+    is_sized_iterable,
+)
 from awkward._slicing import normalize_slice
 from awkward._typing import (
     TYPE_CHECKING,
@@ -38,11 +49,12 @@ from awkward._typing import (
     Protocol,
     Self,
     SupportsIndex,
+    Type,
     TypeAlias,
     TypedDict,
 )
 from awkward._util import UNSET
-from awkward.forms.form import Form
+from awkward.forms.form import Form, FormKeyPathT
 from awkward.index import Index, Index64
 
 if TYPE_CHECKING:
@@ -226,6 +238,12 @@ class Content(Meta):
     ) -> Form:
         raise NotImplementedError
 
+    def form_with_key_path(self, root: FormKeyPathT = ()) -> Form:
+        return self._form_with_key_path(root)
+
+    def _form_with_key_path(self, path: FormKeyPathT) -> Form:
+        raise NotImplementedError
+
     @property
     def form_cls(self) -> type[Form]:
         raise NotImplementedError
@@ -509,10 +527,14 @@ class Content(Meta):
             )
 
     def __getitem__(self, where):
-        return self._getitem(where)
+        return self._getitem(where, NamedAxis)
 
-    def _getitem(self, where):
+    def _getitem(self, where, named_axis: Type[NamedAxis] = NamedAxis):
         if is_integer_like(where):
+            # propagate named_axis to output
+            named_axis.mapping = _remove_named_axis(
+                named_axis.mapping, where, self.purelist_depth
+            )
             return self._getitem_at(ak._slicing.normalize_integer_like(where))
 
         elif isinstance(where, slice) and where.step is None:
@@ -523,21 +545,35 @@ class Content(Meta):
             return self._getitem_range(start, stop)
 
         elif isinstance(where, slice):
-            return self._getitem((where,))
+            return self._getitem((where,), named_axis)
 
         elif isinstance(where, str):
             return self._getitem_field(where)
 
         elif where is np.newaxis:
-            return self._getitem((where,))
+            return self._getitem((where,), named_axis)
 
         elif where is Ellipsis:
-            return self._getitem((where,))
+            return self._getitem((where,), named_axis)
 
         elif isinstance(where, tuple):
             if len(where) == 0:
                 return self
 
+            # count number of ellipsis
+            # Need to use a little trick here:
+            #   where.count(Ellipsis) does not work, because it will do a == comparison against Ellipsis,
+            #   and this will fail in the case of typetracers where == is dispatched to np.equal ufunc.
+            #   In this dispatch we encounter an assertion that the type of the Ellipsis is not allowed.
+            #   ...but luckily we can use the fact that Ellipsis is a singleton and use the 'is' operator
+            n_ellipsis = 0
+            for w in where:
+                if w is ...:
+                    n_ellipsis += 1
+
+            if n_ellipsis > 1:
+                raise IndexError("an index can only have a single ellipsis ('...')")
+
             # Backend may change if index contains typetracers
             backend = backend_of(self, *where, coerce_to_common=True)
             this = self.to_backend(backend)
@@ -547,6 +583,62 @@ class Content(Meta):
             # Prepare items for advanced indexing (e.g. via broadcasting)
             nextwhere = ak._slicing.prepare_advanced_indexing(items, backend)
 
+            # Handle named axis
+            # first expand the ellipsis to colons in nextwhere,
+            # copy nextwhere to not pollute the original
+            _nextwhere = tuple(nextwhere)
+            if n_ellipsis == 1:
+                # collect the ellipsis index
+                # same little trick as above for `nextwhere.index(...)`
+                (ellipsis_at,) = tuple(i for i, x in enumerate(nextwhere) if x is ...)
+                # calculate how many slice(None) we need to add
+                # same little trick as above for `nextwhere.count(None)`
+                n_newaxis = 0
+                for x in nextwhere:
+                    if x is np.newaxis or x is None:
+                        n_newaxis += 1
+                n_total = self.minmax_depth[1]
+                n_slice_none = n_total - (len(_nextwhere) - n_newaxis - 1)
+                # expand `[...]` to `[:]*n_slice_none`
+                _nextwhere = (
+                    _nextwhere[:ellipsis_at]
+                    + (slice(None),) * n_slice_none
+                    + _nextwhere[ellipsis_at + 1 :]
+                )
+
+            # now propagate named axis
+            _named_axis = _keep_named_axis(named_axis.mapping, None)
+            _adjust_dim = 0
+            # this loop does the following:
+            # - remove a named axis for integer indices, e.g. `a[1, 2]`
+            # - add a named axis for None (or np.newaxis) indices, e.g. `a[..., None]`
+            # - keep named axis for any other index, e.g. `a[:]`, `a[0:1]`, or `a[a>0]`
+            #   (these may only remove elements, but not dimensions)
+            for dim, nw in enumerate(_nextwhere):
+                dim_adjusted = dim + _adjust_dim
+                total_adjusted = self.minmax_depth[1] + _adjust_dim
+                for _, pos in _named_axis.items():
+                    if maybe_posaxis(self, pos, 0) == dim_adjusted:
+                        break
+
+                if is_integer(nw) or (is_array_like(nw) and nw.ndim == 0):
+                    _named_axis = _remove_named_axis(
+                        named_axis=_named_axis,
+                        axis=dim_adjusted,
+                        total=total_adjusted,
+                    )
+                    _adjust_dim -= 1
+                elif nw is None:
+                    _named_axis = _add_named_axis(
+                        named_axis=_named_axis,
+                        axis=dim_adjusted,
+                        total=total_adjusted,
+                    )
+                    _adjust_dim += 1
+
+            # set propagated named axis
+            named_axis.mapping = _named_axis
+
             next = ak.contents.RegularArray(
                 this,
                 this.length,
@@ -562,7 +654,7 @@ class Content(Meta):
                 return out._getitem_at(0)
 
         elif isinstance(where, ak.highlevel.Array):
-            return self._getitem(where.layout)
+            return self._getitem(where.layout, named_axis)
 
         # Convert between nplikes of different backends
         elif (
@@ -570,7 +662,9 @@ class Content(Meta):
             and where.backend is not self._backend
         ):
             backend = backend_of(self, where, coerce_to_common=True)
-            return self.to_backend(backend)._getitem(where.to_backend(backend))
+            return self.to_backend(backend)._getitem(
+                where.to_backend(backend), named_axis
+            )
 
         elif isinstance(where, ak.contents.NumpyArray):
             data_as_index = to_nplike(
@@ -602,7 +696,7 @@ class Content(Meta):
                     allow_lazy = "copied"  # True, but also can be modified in-place
                 else:
                     wheres = self._backend.index_nplike.nonzero(data_as_index)
-                    return self._getitem(wheres)
+                    return self._getitem(wheres, named_axis)
             else:
                 raise TypeError(
                     "array slice must be an array of integers or booleans, not\n\n    {}".format(
@@ -621,9 +715,9 @@ class Content(Meta):
         elif isinstance(where, ak.contents.RegularArray):
             maybe_numpy = where.maybe_to_NumpyArray()
             if maybe_numpy is None:
-                return self._getitem((where,))
+                return self._getitem((where,), named_axis)
             else:
-                return self._getitem(maybe_numpy)
+                return self._getitem(maybe_numpy, named_axis)
 
         # Awkward Array of strings
         elif (
@@ -637,7 +731,7 @@ class Content(Meta):
             return where.to_NumpyArray(np.int64)
 
         elif isinstance(where, Content):
-            return self._getitem((where,))
+            return self._getitem((where,), named_axis)
 
         elif is_sized_iterable(where):
             # Do we have an array
@@ -654,7 +748,7 @@ class Content(Meta):
                     primitive_policy="error",
                     string_policy="as-characters",
                 )
-                return self._getitem(layout)
+                return self._getitem(layout, named_axis)
 
             elif len(where) == 0:
                 return self._carry(
@@ -669,17 +763,20 @@ class Content(Meta):
                 return self._getitem_fields(list(where))
 
             else:
-                layout = ak.operations.ak_to_layout._impl(
-                    where,
-                    allow_record=False,
-                    allow_unknown=False,
-                    none_policy="error",
-                    regulararray=False,
-                    use_from_iter=True,
-                    primitive_policy="error",
-                    string_policy="as-characters",
+                layout = ak.to_backend(
+                    ak.operations.ak_to_layout._impl(
+                        where,
+                        allow_record=False,
+                        allow_unknown=False,
+                        none_policy="error",
+                        regulararray=False,
+                        use_from_iter=True,
+                        primitive_policy="error",
+                        string_policy="as-characters",
+                    ),
+                    self._backend,
                 )
-                return self._getitem(layout)
+                return self._getitem(layout, named_axis)
 
         else:
             raise TypeError(
@@ -693,6 +790,9 @@ class Content(Meta):
     def _is_getitem_at_placeholder(self) -> bool:
         raise NotImplementedError
 
+    def _is_getitem_at_virtual(self) -> bool:
+        return NotImplementedError
+
     def _getitem_at(self, where: IndexType):
         raise NotImplementedError
 
@@ -1007,6 +1107,10 @@ class Content(Meta):
     ):
         raise NotImplementedError
 
+    def _to_cudf(self, cudf: Any, mask: Content | None, length: int):
+        # prototype abstract signature
+        raise NotImplementedError
+
     def to_backend_array(
         self, allow_missing: bool = True, *, backend: Backend | str | None = None
     ):
@@ -1202,6 +1306,28 @@ class Content(Meta):
     def _to_backend(self, backend: Backend) -> Self:
         raise NotImplementedError
 
+    def materialize(self) -> Self:
+        return self._materialize()
+
+    def _materialize(self) -> Self:
+        raise NotImplementedError
+
+    @property
+    def is_all_materialized(self) -> bool:
+        return self._is_all_materialized
+
+    @property
+    def _is_all_materialized(self) -> bool:
+        raise NotImplementedError
+
+    @property
+    def is_any_materialized(self) -> bool:
+        return self._is_any_materialized
+
+    @property
+    def _is_any_materialized(self) -> bool:
+        raise NotImplementedError
+
     def with_parameter(self, key: str, value: Any) -> Self:
         out = copy.copy(self)
 
@@ -1261,6 +1387,10 @@ class Content(Meta):
     def copy(self, *, parameters: JSONMapping | None = UNSET) -> Self:
         raise NotImplementedError
 
+    @classmethod
+    def _arrow_needs_option_type(cls):
+        return cls.is_option  # is_option is a class property of Meta
+
 
 @register_backend_lookup_factory
 def find_content_backend(obj: type):
